#!/usr/bin/python
# Copyright 2016 CERN and GSI
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import libxml2
from collections import namedtuple

import iecommon
import iefiles
import genduwrappertemplate

from iecommon import *

#=========================================================================
# Sub-functions
#=========================================================================
    
def extractBlockList(designDOM):
    blockList = []
    for blockNode in designDOM.xpathEval("//Block"):
        blockList.append([blockNode.prop("name"), blockNode.prop("mode")])
    return blockList

def getRegisterList(designFile,blockName):
    '''Return the list of register included in a given block.
    for each register a the following tuple is returned: [name,format,dim1,dim2]
    '''
    registerList = []
    dom = libxml2.parseFile(designFile)
    blockNodeList = dom.xpathEval("//Block")
    for blockNode in blockNodeList:
        if blockNode.prop("name")==blockName:
            break #after the loop blockNode will point to the good node
        
    registerNodeList=blockNode.xpathEval("Register")
    for registerNode in registerNodeList:
        name = registerNode.prop("name")
        format = registerNode.prop("format")
        dim1 = 1
        dim2 = 1
        try:
            dim1 = int(registerNode.prop("array-dim1"))
        except:
            pass #use 1 as default dimention 
        try:
            dim2 = int(registerNode.prop("array-dim2"))
        except:
            pass #use 1 as default dimention
        registerList.append([name,format,dim1,dim2])
    return registerList
            
def genClassHeader(workspacePath, deployName, classNode, funcGetSilecsDesignFilePath, funcGetDuDesignWrapperFile, logTopics):
    designName = classNode.prop("silecs-design-name")
    designVersion = classNode.prop("silecs-design-version") 
    designFile = funcGetSilecsDesignFilePath(workspacePath, designName)
    designDOM = libxml2.parseFile(designFile)
    if(not os.path.isfile(designFile)):
        raise Exception("File not found: " + designFile)
    blockList = extractBlockList(designDOM)
    classDeclarations = ""
    
    #construct Block class
    for [blockName,blockMode] in blockList:
        registerList = getRegisterList(designFile,blockName)
        registerInitializerList = genduwrappertemplate.getBlockInitializerList(registerList)
        registerGetterSetter = genduwrappertemplate.getRegisterGetterSetter(blockName,blockMode,registerList)
        registersDimentionsDeclaration = genduwrappertemplate.getRegistersDimetionsDeclaration(registerList)
        registersDeclaration = genduwrappertemplate.getRegistersDeclaration(registerList)
        classDeclarations += genduwrappertemplate.getBlockClass(blockName,registerInitializerList,registerGetterSetter,registersDimentionsDeclaration,registersDeclaration)
    
    blockGetters = ""
    sendRecvBlocks = genduwrappertemplate.getDeviceSendRecvBlocks(blockList)
    for [blockName,mode] in blockList:
        registerList = getRegisterList(designFile,blockName)
        blockGetters += genduwrappertemplate.getDeviceBlockGetterSetter(blockName, mode, registerList)

    classDeclarations += genduwrappertemplate.getDeviceClass(blockGetters,sendRecvBlocks)

    sendRecvBlocks = genduwrappertemplate.getControllerSendRecvBlocks(blockList)
    classDeclarations = genduwrappertemplate.getDesignClass(designName, designVersion)
    designWrapper = genduwrappertemplate.designFileTemplate.substitute({'designNameCapitalized' :  iecommon.capitalizeString(designName),'designNameUpper' :  designName.upper(),'classDeclarations' : classDeclarations})

    designWrapperFile = funcGetDuDesignWrapperFile(workspacePath, deployName, designName)
    fdesc = open(designWrapperFile, "w")
    fdesc.write(designWrapper)
    iecommon.logInfo('Generated Wrapper for Design %s'%(designName), logTopics)
    fdesc.close()

def genDuWrapperBase(deployFile,funcGetDuWrapperFile,workspacePath,funcGetSilecsDesignFilePath, funcGetDuDesignWrapperFile,deployName,deployVersion,logTopics={'errorlog': True}):
    deployDOM = libxml2.parseFile(deployFile)
    
    for controllerNode in deployDOM.xpathEval("/SILECS-Deploy/Controller"):
        controllerName = controllerNode.prop("host-name")
        controllerDomain = "" # GSI-Hack - No Support for domains at GSI
        deployCustomInclude = constructorBody = destructorBody = designGetters = designMemberDeclarations = ""
        for classNode in controllerNode.xpathEval("SilecsDesign"):
            designName = classNode.prop("silecs-design-name")
            designNameCapitalized = iecommon.capitalizeString(designName)
            genClassHeader(workspacePath, deployName, classNode, funcGetSilecsDesignFilePath, funcGetDuDesignWrapperFile, logTopics)
            deployCustomInclude += genduwrappertemplate.deployIncludeTemplate.substitute({'designNameCapitalized' : designNameCapitalized})
            constructorBody += genduwrappertemplate.designAllocation.substitute({'designName' : designName,'designNameCapitalized' : designNameCapitalized})
            destructorBody += genduwrappertemplate.designDeallocation.substitute({'designName' : designName})
            designGetters += genduwrappertemplate.designGetterTemplate.substitute({'designName' : designName,'designNameCapitalized' : designNameCapitalized})
            designMemberDeclarations += genduwrappertemplate.deployUnitMembersDeclaration.substitute({'designName' : designName,'designNameCapitalized' : designNameCapitalized})

        controllerCtor = deviceGetter = ''
        for deviceNode in controllerNode.xpathEval("*/Device"):
            deviceName = deviceNode.prop("device-name")
            controllerCtor +=  genduwrappertemplate.controllerDeviceInit.substitute({'deviceName' : deviceName})
            deviceGetter += genduwrappertemplate.deviceGetterTemplate.substitute({'deviceName' : deviceName, 'deviceNameCapitalizedNoUndercore' : iecommon.capitalizeString(deviceName.replace("-","_"))})
        controllerClass = genduwrappertemplate.getControllerClass(controllerCtor,deviceGetter,controllerName, controllerDomain)
        code = genduwrappertemplate.generateControllerFile(deployName,deployVersion,deployCustomInclude,constructorBody,destructorBody,designGetters,designMemberDeclarations,controllerClass)
    
        fdesc = open(funcGetDuWrapperFile(workspacePath, deployName, controllerNode.prop("host-name")) , "w")
        fdesc.write(code)
        iecommon.logInfo('Generated Wrapper for Controller %s'%(controllerNode.prop("host-name")), logTopics)
        fdesc.close()

def genDuWrapper(workspacePath,deployName,deployVersion,logTopics={'errorlog': True}):
    deployFile = iefiles.getSilecsDeployFilePath(workspacePath, deployName)
    funcGetDuWrapperFile = iefiles.getDuWrapperFile
    
    # Create output directory if necessary
    sourcePath = iefiles.getDuWrapperSourceDirectory(workspacePath, deployName)
    if not os.path.exists(sourcePath):
        os.makedirs(sourcePath)
        
    genDuWrapperBase(deployFile,funcGetDuWrapperFile,workspacePath,iefiles.getSilecsDesignFilePath, iefiles.getDuDesignWrapperFile,deployName,deployVersion,logTopics)
    
    
