#!/usr/bin/python
# Copyright 2016 CERN and GSI
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import time
import zlib
import libxml2

import iecommon
import iefiles
import xmltemplate
import s7template
import virtualS7Template
import rabbitTemplate

from iecommon import *
#=========================================================================
# General remarks
#=========================================================================

#XML input documents well-formed is not checked! 
#This Python-script generates PLC sources document from XML Parameter documents
#which are supposed to be well-formed (generated with the 'genparam' script).


# Deploy global information
class Deploy(object):
    plcModel        = None  # S7-300, .., Premium, ..
    plcSystem       = None  # UNITY, STEP-7, TIA-PORTAL, ...
    plcProtocol     = None  # DEVICE_MODE or BLOCK_MODE
    address         = None
    checksum     = None  # checksum of the related parameter file
    owner           = None  # onwer of the Deployment document
    silecsVersion   = None

whichUnityFormat = {
    'uint8'      : 'BYTE',
    'int8'        : 'WORD',
    'uint16'    : 'WORD',
    'int16'      : 'INT',
    'uint32'    : 'DWORD',
    'int32'      : 'DINT',
    'float32'   : 'REAL',
    'string'     : 'STRING',
#    'uint64'      : '8',    not supported with PLC
#    'int64'        : '8',    not supported with PLC
#    'float64'     : '8',    not supported with PLC
    'date'       : 'DT',
    'char'       : 'WORD',
    'byte'       : 'BYTE',
    'word'      : 'WORD',
    'dword'    : 'DWORD',
    'int'          : 'INT',
    'dint'        : 'DINT',
    'real'        : 'REAL',
    'dt'           : 'DT'
}

whichTwincatFormat = {
    'uint8'      : 'BYTE',
    'int8'        : 'SINT',
    'uint16'    : 'WORD',
    'int16'      : 'INT',
    'uint32'    : 'DWORD',
    'int32'      : 'DINT',
    'float32'   : 'REAL',
    'string'     : 'STRING',
#    'uint64' not supported by PLCs
#    'int64' not supported by PLCs
#    'float64' not supported by PLCs
    'date'       : 'DT',
    'char'       : 'SINT',
    'byte'       : 'BYTE',
    'word'      : 'WORD',
    'dword'    : 'DWORD',
    'int'          : 'INT',
    'dint'        : 'DINT',
    'real'        : 'REAL',
    'dt'           : 'DT',
    'string'     : 'STRING'
}

schneiderRegArray  = """ARRAY[0..%d] OF %s"""
schneiderRegArray2d  = """ARRAY[0..%d, 0..%d] OF %s"""

# Beckhoff templates (using newlines \r\n because TwinCAT compiler complains with line feed character)
beckhoffReg = """    %s_%04x_%s AT %%MW%s: %s;\r\n\r\n"""
beckhoffRegArray = """    %s_%04x_%s AT %%MW%s: ARRAY [0..%d] OF %s;\r\n\r\n"""
beckhoffRegArray2d = """    %s_%04x_%s AT %%MW%s: ARRAY [0..%d, 0..%d] OF %s;\r\n\r\n"""
beckhoffRegInit = """    %s_%04x_%s AT %%MW%s: %s:= %s;\r\n\r\n"""

#=========================================================================
# Sub-function
#=========================================================================

#-------------------------------------------------------------------------
# Schneider registers
#-------------------------------------------------------------------------

def xsyRegister(regFormat, regDim, regDim2, regLen=1):
    if regLen > 1:  # register is a string or an array of strings
        strLen= whichUnityFormat[regFormat]+'['+str(regLen)+']'
        if  regDim == 1 and regDim2 ==1:    # string
            return strLen
        elif regDim > 1 and regDim2 == 1:   # string array
            return schneiderRegArray %(regDim-1, strLen)
        else: # dim1>=1 and for whatever dim2, use double array syntax
            return schneiderRegArray2d %(regDim-1, regDim2-1, strLen)
    else:
        if  regDim == 1 and regDim2 ==1:    # scalar
            return whichUnityFormat[regFormat]
        elif regDim > 1 and regDim2 == 1:   # array
            return schneiderRegArray %(regDim-1, whichUnityFormat[regFormat])
        else: # dim1>=1 and for whatever dim2, use double array syntax
            return schneiderRegArray2d %(regDim-1, regDim2-1, whichUnityFormat[regFormat])
    
def getDateTime():
    dt = time.localtime(time.time())
    return "DT#%s-%s-%s-%s:%s:%s" %(dt[0],dt[1],dt[2],dt[3],dt[4],dt[5])

#-------------------------------------------------------------------------
# SIEMENS PLC code generation
#-------------------------------------------------------------------------
def generateSiemensSources(paramDOM, sourceFolderPath ,logTopics):
    stlString = ''
    symString = ''  # for .sdf Symbol file

    deploy = getDeploymentInformation(paramDOM)

    # Prepare the Simatic Symbols (<.sdf> file)
    DBnumber    = long(deploy.address)       #first data-block (SilecsHeader) starts from the base-address
    UDTnumber   = long(deploy.address)       #use base data-block address for UDT number as well (to be reserved by the developer)

    # Generate sources for each class that is deployed in the PLC
    classNodes = paramDOM.xpathEval("/SILECS-Param/SILECS-Mapping/SILECS-Class")
    for classNode in classNodes:
        blockDOMList  = classNode.xpathEval("Block") # get Blocks elements of that class
        deviceDOMList = classNode.xpathEval("Instance") # get Device instances of that class
        
        className       = classNode.prop('name')
        classVersion    = classNode.prop('version')

        deviceNumber = len(deviceDOMList)

        # Generate the Blocks definition
        for blockIndex, blockDOM in enumerate(blockDOMList):
            registerList = ''
            blockName = blockDOM.prop('name')
            registerDOMList = blockDOM.xpathEval("Register") # get registers of the current block
            for registerIndex, registerDOM in enumerate(registerDOMList):
                registerName = registerDOM.prop('name')
                registerFormat = registerDOM.prop('format')
                
                # If PLC does not supports this register format abort this generation and log the error
                if(registerFormat in ['uint64','int64','float64']):
                    iecommon.logError('ERROR: In design %s_%s register %s, %s format not supported for current controller model.'
                                      %(className, classVersion, registerName, registerFormat)
                                      , True,logTopics)
                
                #create register value if required (diagnostic registers assignment in particular)
                registerValue = s7template.stlRegisterValue(registerName, deploy.owner, getDateTime(), deploy.checksum,deploy.silecsVersion)
                
                # Read register dimensions
                registerDimension1 = long(registerDOM.prop('array-dim1'))
                registerDimension2 = long(registerDOM.prop('array-dim2'))
                
                if registerFormat == 'string':
                    registerList += s7template.stlRegister(registerName, registerFormat, registerDimension1, registerDimension2, registerValue, long(registerDOM.prop('string-len')))
                else:                    
                    registerList += s7template.stlRegister(registerName, registerFormat, registerDimension1, registerDimension2, registerValue)

            stlString += s7template.stlBlockUDT(deploy.owner, className, classVersion, blockName, registerList, (blockIndex == 0))
            symString += s7template.symBlockUDT(className, classVersion, blockName, UDTnumber, (blockIndex == 0))
            UDTnumber += 1
   
        if deploy.plcProtocol == 'DEVICE_MODE':
        # Generate the Data-Blocks: one DB per Device instance
            for deviceIndex, deviceDOM in enumerate(deviceDOMList):
                blockList = ''
                for blockDOM in blockDOMList:
                    blockName = blockDOM.prop('name')
                    blockList += s7template.stlBlock(className, blockName)
                deviceLabel = deviceDOM.prop('label')
                stlString += s7template.generateBlock(deploy.owner, className, classVersion, deploy.plcSystem, DBnumber, deviceLabel, blockList, (deviceIndex == 0),deploy.plcProtocol)
                symString += s7template.symDeviceDB(className, classVersion, deviceLabel, DBnumber, (deviceIndex == 0))
                DBnumber += 1

        else:             # BLOCK_MODE
        # Generate the Data-Blocks: one DB per Block of registers
            for blockIndex, blockDOM in enumerate(blockDOMList):
                blockName = blockDOM.prop('name')

                deviceList = ''
                for deviceDOM in deviceDOMList:
                    deviceLabel = deviceDOM.prop('label')
                    deviceList += s7template.stlDevice(deviceLabel, className, blockName)

                stlString += s7template.generateBlock(deploy.owner, className, classVersion, deploy.plcSystem, DBnumber, blockName, deviceList, (blockIndex == 0),deploy.plcProtocol)
                symString += s7template.symBlockDB(className, classVersion, blockName, DBnumber, (blockIndex == 0))
                DBnumber += 1

    # write the String PLC generated code
    generateControllerFiles(sourceFolderPath,deploy,".scl",stlString,logTopics);
        
    # Write the String PLC generated symbols into the <.sdf> symbols file
    generateControllerFiles(sourceFolderPath,deploy,".sdf",symString,logTopics);

#-------------------------------------------------------------------------
# SCHNEIDER PLC code generation
#-------------------------------------------------------------------------

# ----------------------------------------------------
def generateSchneiderRegisters(xsydoc, classDOM, deviceDOM, blockDOM, deviceIndex, modeDevice, deploy, logTopics):
    
    className    = classDOM.prop('name')
    deviceLabel = deviceDOM.prop('label')
    blockName    = blockDOM.prop('name')
    
    #Device and Block adresses come from the generated Paramerers file. Depends on the protocol mode:
    #DEVICE_MODE: relative address of the element within it parent node
    #BLOCK_MODE : absolute address in the entire PLC memory
    deviceAddress = long(deviceDOM.prop('address'))
    blockAddress  = long(blockDOM.prop('address'))
    blockMemSize  = long(blockDOM.prop('mem-size'))
    registerDOMList = blockDOM.xpathEval("Register") # get Registers of the current block
    for registerIndex,registerDOM in enumerate(registerDOMList):
        registerName = registerDOM.prop('name')
        registerFormat = registerDOM.prop('format')
        registerDimension1 = long(registerDOM.prop('array-dim1'))
        registerDimension2 = long(registerDOM.prop('array-dim2'))

        # If PLC does not supports this register format abort this generation and log the error
        if(registerFormat in ['uint64','int64','float64']):
            iecommon.logError('ERROR: In register '+registerName+', '+registerFormat+' format not supported for current controller model.', True,logTopics)
        
        #register has relative address irrespective of the protocol mode.
        registerOffset = long(registerDOM.prop('address'))
        
        # Compute the Register address relying on the Class Parameters file data
        # Attention! Schneider uses WORD addressing while Parameters data are expressed in bytes to be PLC independent
        if modeDevice:
            registerAddress = (deviceAddress + blockAddress + registerOffset)/2
        else:
            registerAddress = (blockAddress + (deviceIndex * blockMemSize) + registerOffset)/2
              
        # then insert the corresponding DOM element
        dataElt = xsydoc.xpathEval("//dataBlock")[0] #only one dataBlock node has been inserted so far
        regElt  = libxml2.newNode('variables')
        # Compute the base checksum of the class name
        classNameCRC = zlib.crc32(className,0) & 0xffff
        # Create register name = regname_crc32(classname)_deviceid
        # 24 char (register name) + 1 (underscore) + 4 char (CRC classname) + 1 (underscore) + n (device id) = 30 + n (device id) 
        # possible problem when device id (n)> 99 --> string > 32 not compatible with Schneider PLCs
        regElt.setProp("name", "%s_%04x_%s" %(registerName, classNameCRC, deviceLabel))
        
        if registerFormat == 'string':
            regElt.setProp("typeName", "%s" %xsyRegister(registerFormat, registerDimension1, registerDimension2, registerDOM.prop('string-len')))
        else:
            regElt.setProp("typeName", "%s" %xsyRegister(registerFormat, registerDimension1, registerDimension2))
        
            
        regElt.setProp("topologicalAddress", "%%MW%ld" %registerAddress)
        if blockName == 'hdrBlk':
            initElt = libxml2.newNode('variableInit')
            if registerName == '_version': initElt.setProp("value", deploy.silecsVersion )
            if registerName == '_user'   : initElt.setProp("value", deploy.owner)
            if registerName == '_checksum': initElt.setProp("value", "%s" %deploy.checksum)
            if registerName == '_date'  : initElt.setProp("value", getDateTime())
            regElt.addChild(initElt)
        commElt    = libxml2.newNode('comment')
        commElt.setContent(className+"/"+deviceLabel+"/"+blockName)
        regElt.addChild(commElt)
        dataElt.addChild(regElt)
        
# ----------------------------------------------------
def generateSchneiderSources(paramDOM,sourceFolderPath,logTopics):
    #Base address provided by the user in the Deployment relies on the PLC model:  
    #Quantum, Premium and M340 Schneider PLCs are all interpreted as using 16-bit addressing, 
    #even though M340 addresses are later properly considered as 32-bit addresses
    
    deploy = getDeploymentInformation(paramDOM)

    # Create the Schneider DOM source (Unity <.XSY> XML document)
    xsyDoc  = libxml2.newDoc(version='1.0')
    rootElt = libxml2.newNode('VariablesExchangeFile')
    xsyDoc.addChild(rootElt)
    dataElt = libxml2.newNode('dataBlock')
    rootElt.addChild(dataElt)
   
    classNodes = paramDOM.xpathEval("/SILECS-Param/SILECS-Mapping/SILECS-Class")
    for classNode in classNodes:
        address     = int(classNode.prop('address')) #address of the first register of the configuration
        blockDOMList  = classNode.xpathEval("Block") # get Blocks elements of that class
        deviceDOMList = classNode.xpathEval("Instance") # get Device instances of that class
        
        for blockDOM in blockDOMList:
            for deviceIndex,deviceDOM in enumerate(deviceDOMList):
                if deploy.plcProtocol == 'DEVICE_MODE': #-------------------
                    generateSchneiderRegisters(xsyDoc, classNode, deviceDOM, blockDOM, deviceIndex, True, deploy, logTopics)
                else:            
                    generateSchneiderRegisters(xsyDoc, classNode, deviceDOM, blockDOM, deviceIndex, False, deploy, logTopics)
    
    # Finally, write the DOM object (PLC generated code) into the XSY source file
    generateControllerFiles(sourceFolderPath,deploy,".xsy",rootElt.serialize(format = True),logTopics);

    xsyDoc.freeDoc()      
    
# ----------------------------------------------------
def generateRabbitSources(paramDOM,sourceFolderPath,logTopics):   
    deviceLabel = ''

    deploy = getDeploymentInformation(paramDOM)

    # Prepare the source (<.h> file) with its diagnostic data-block header
    cCodeString =  rabbitTemplate.cHeader(deploy.address,deploy.silecsVersion)
    cTypeDefinitions = ''
    cDataAllocation  = rabbitTemplate.cAllocationComment(deploy.plcProtocol)
    
    blockList = ''
    classList = ''
    NBdeviceDefinitionList = ''

    classNodes = paramDOM.xpathEval("/SILECS-Param/SILECS-Mapping/SILECS-Class")
    for classNode in classNodes:
        blockDOMList  = classNode.xpathEval("Block")    # get Blocks elements of that class
        deviceDOMList = classNode.xpathEval("Instance") # get Device instances of that class
        
        className        = classNode.prop('name')
        classVersion    = classNode.prop('version')

        deviceNumber = len(deviceDOMList)

        #==== Generate the Blocks definition ====
        for blockIndex, blockDOM in enumerate(blockDOMList):
            registerList = ''
            blockName = blockDOM.prop('name')
            registerDOMList = blockDOM.xpathEval("Register") # get Registers of the current block
            for registerIndex, registerDOM in enumerate(registerDOMList):
                registerName = registerDOM.prop('name')
                registerFormat = registerDOM.prop('format')
                registerDimension1 = long(registerDOM.prop('array-dim1'))
                registerDimension2 = long(registerDOM.prop('array-dim2'))
                
                # If PLC does not supports this register format abort this generation and log the error
                if(registerFormat in ['float64']):
                    iecommon.logError('ERROR: In register '+registerName+', '+registerFormat+' format not supported for current controller model.', True,logTopics)
                
                if(registerFormat == 'string'):
                    registerList += rabbitTemplate.cRegister(registerName, registerFormat, registerDimension1, registerDimension2, long(registerDOM.prop('string-len')))
                else:
                    registerList += rabbitTemplate.cRegister(registerName, registerFormat, registerDimension1, registerDimension2)    

            cTypeDefinitions += rabbitTemplate.cBlockUDT(className, classVersion, blockName, registerList, (blockIndex == 0))
            
        #==== Memory allocation ====
        if deploy.plcProtocol == 'DEVICE_MODE':
            # Generate the List of block in the class
            blockList =''
            for blockIndex, blockDOM in enumerate(blockDOMList):
                blockName = blockDOM.prop('name')
                blockList += rabbitTemplate.cDeviceModeBlockInstantiation(className, blockName)

            deviceList = ''
            for deviceDOM in deviceDOMList:
                deviceLabel = className + "_" + deviceDOM.prop('label')
                deviceList += deviceLabel + ', ';
            deviceList = deviceList[:-2]  # remove last comma space
            classList += rabbitTemplate.cDeviceModeClass_deviceList(blockList, deviceList)


        else:  # BLOCK_MODE
            # Generate the List of block in the class
            for blockIndex, blockDOM in enumerate(blockDOMList):
                # create instantation for current block
                blockName = blockDOM.prop('name')

                deviceList = ''
                for deviceDOM in deviceDOMList:
                    deviceLabel = deviceDOM.prop('label')
                    deviceList +=rabbitTemplate.cBlockModeDeviceInstantiation_deviceList(className, blockName, deviceLabel)

                blockList += rabbitTemplate.cBlockModeBlockInstantiation(deviceList, className, blockName)

    # Predefine number of devices constant
    cDataAllocation += NBdeviceDefinitionList                

    if deploy.plcProtocol == 'DEVICE_MODE':
        cDataAllocation += rabbitTemplate.cDeviceModeDataInstantiation(classList)
        cInitFunction    = rabbitTemplate.cInitDeviceMode(deploy.silecsVersion,deploy.checksum,deploy.owner)
    else:
        cDataAllocation += rabbitTemplate.cBlockModeDataInstantiation(blockList)
        cInitFunction    = rabbitTemplate.cInitBlockMode(deploy.silecsVersion,deploy.checksum,deploy.owner)
    
            
    # Append data definitions
    cCodeString += cTypeDefinitions
    
    # Append data allocation
    cCodeString += cDataAllocation
    
    # Append initialization function
    cCodeString += cInitFunction
    
    # Generate and append sample example
    cCodeString += rabbitTemplate.cExample(deploy.plcProtocol,True,className, blockName, registerName, deviceLabel)
    
    # Finally, write the String C generated code into the temporal output file
    generateControllerFiles(sourceFolderPath,deploy,".h",cCodeString,logTopics);


# ----------------------------------------------------
def generateVirtualS7Sources(paramDOM,sourceFolderPath,logTopics):   
    deploy = getDeploymentInformation(paramDOM)
    includeDesign  = ''
    allocDesign  = ''
    deleteDesign = ''
    getDesign = ''

    protocolMode = 'BlockMode'; 
    if deploy.plcProtocol == 'DEVICE_MODE':
        protocolMode = 'DeviceMode';
    duConstructor = virtualS7Template.vs7DuConstructor(deploy.plcName, protocolMode, deploy.address)
    
    classNodes = paramDOM.xpathEval("/SILECS-Param/SILECS-Mapping/SILECS-Class")
    
    # Generate sources for each class that is deployed in the PLC
    for classNode in classNodes:
    
        blocksCodeString = ''
         
        createBlockCodeString = '' 
        deleteBlockCodeString = '' 
        deviceCodeString = ''
         
        createDeviceCodeString = '' 
        deleteDeviceCodeString = '' 
        designCodeString = ''
         
        blockDOMList  = classNode.xpathEval("Block")    # get Blocks elements of that class
        deviceDOMList = classNode.xpathEval("Instance") # get Device instances of that class
        
        className = classNode.prop('name')
        classVersion = classNode.prop('version')
    
        #DEPLOY-UNIT code generation ============================
        includeDesign += virtualS7Template.vs7DuDesignInclude(className, classVersion)   
        allocDesign += virtualS7Template.vs7DuDesignAlloc(className, classVersion)   
        deleteDesign += virtualS7Template.vs7DuDesignDelete(className)
        getDesign += virtualS7Template.vs7DuDesignGet(className, classVersion)

        #CLASSES code generation ================================
        for blockIndex, blockDOM in enumerate(blockDOMList):
            getSetCodeString = ''
            dimsCodeString   = ''
            dataCodeString   = ''
            
            currentRegisterAddress = 0  #used to compute even/odd adress and insert align data by adding dummy registers
            previousRegisterMemSize = 0 #...
            dummyIndex = 0;
             
            blockName = blockDOM.prop('name')
            blockOffset = blockDOM.prop('address')
            registerDOMList = blockDOM.xpathEval("Register") # get Registers of the current block
            
            createBlockCodeString += virtualS7Template.vs7ClassCreateBlock(className, blockName)
            deleteBlockCodeString += virtualS7Template.vs7ClassDeleteBlock(className, blockName)
            
            for registerIndex, registerDOM in enumerate(registerDOMList):
                registerName = registerDOM.prop('name')
                registerFormat = registerDOM.prop('format')
                registerDimension1 = long(registerDOM.prop('array-dim1'))
                registerDimension2 = long(registerDOM.prop('array-dim2'))
                registerMemSize = long(registerDOM.prop('mem-size')) #store the reigster mem-size for next iteration
                
                registerLength = 1  #in case string 
                if(registerFormat == 'string'):
                    registerLength = long(registerDOM.prop('string-len'))
            
                getSetCodeString += virtualS7Template.vs7ClassGetSet(registerName, registerFormat, registerDimension1, registerDimension2,registerLength)
                dimsCodeString += virtualS7Template.vs7ClassDimension(registerName, registerFormat, registerDimension1, registerDimension2,registerLength)

                #Specific code to force 16bit alignment if not scalar 8bit register (==> insert dummy register if needed)
                dummyCodeString = virtualS7Template.vs7ClassDummyRegister(currentRegisterAddress, previousRegisterMemSize, registerMemSize, dummyIndex)
                dataCodeString += dummyCodeString
                currentRegisterAddress += registerMemSize 
                if dummyCodeString != "":
                    dummyIndex += 1
                    currentRegisterAddress += 1     
                #and now insert the normal data register     
                dataCodeString += virtualS7Template.vs7ClassDataRegister(registerName, registerFormat, registerDimension1, registerDimension2, registerLength)
                previousRegisterMemSize = registerMemSize 
                
            blocksCodeString += virtualS7Template.vs7ClassBlock(deploy.silecsVersion, deploy.owner, deploy.checksum, className, blockName, getSetCodeString, blockOffset, dimsCodeString, dataCodeString)

        for deviceIndex, deviceDOM in enumerate(deviceDOMList):
            deviceLabel = deviceDOM.prop('label')
            createDeviceCodeString += virtualS7Template.vs7ClassCreateDevice(className, deviceLabel, deviceIndex)
            deleteDeviceCodeString += virtualS7Template.vs7ClassDeleteDevice(className, deviceLabel)
        
        deviceCodeString = virtualS7Template.vs7ClassDevice(createBlockCodeString, deleteBlockCodeString)
        designCodeString = virtualS7Template.vs7ClassDesign(className, classVersion, createDeviceCodeString, deleteDeviceCodeString)
        
        classCodeString = virtualS7Template.vs7ClassHeader(className, classVersion, blocksCodeString, deviceCodeString, designCodeString)
            
        # write the CLASS generated code into the temporal output file
        generateControllerFiles(sourceFolderPath,deploy, "." + className + ".h",classCodeString,logTopics);
    
    # Prepare the source (<.h> file) with its diagnostic data-block header
    duCodeString = virtualS7Template.vs7DuHeader(deploy.plcName, duConstructor, includeDesign, allocDesign, deleteDesign, getDesign)    
             
    # Write the DEPLOY-UNIT generated code into the temporal output file
    generateControllerFiles(sourceFolderPath,deploy,".h",duCodeString,logTopics);
             
    # Prepare the source (<.h> file) with its diagnostic data-block header
    mainCodeString = virtualS7Template.vs7MainCode(deploy.plcName)
        
    # write the DEPLOY-UNIT generated code into the temporal output file
    generateControllerFiles(sourceFolderPath,deploy,".cpp",mainCodeString,logTopics);
    
# ----------------------------------------------------
def generateBeckhoffRegisters(classDOM, deviceDOM, blockDOM, deviceIndex, deploy, logTopics):
    source = ''    
    className    = classDOM.prop('name')
    deviceLabel = deviceDOM.prop('label')
    blockName    = blockDOM.prop('name')
    
    #Device and Block adresses come from the generated Paramerers file. Depends on the protocol mode:
    #DEVICE_MODE: relative address of the element within it parent node
    #BLOCK_MODE : absolute address in the entire PLC memory
    deviceAddress = long(deviceDOM.prop('address'))
    blockAddress  = long(blockDOM.prop('address'))
    blockMemSize  = long(blockDOM.prop('mem-size'))

    registerDOMList = blockDOM.xpathEval("Register") # get registers of the current block
    
    for registerIndex, registerDOM in enumerate(registerDOMList):
        registerName = registerDOM.prop('name')
        registerFormat = registerDOM.prop('format')
        registerDimension1 = long(registerDOM.prop('array-dim1'))
        registerDimension2 = long(registerDOM.prop('array-dim2'))
        
        if registerDOM.prop('string-len'):  # only appears for string registers
            registerLength = long(registerDOM.prop('string-len'))
        
        if blockName == 'hdrBlk':
            if registerName == 'ieVersion' or registerName == 'ieUser':
                registerFormat = 'string'
                registerDimension1 = 1
    
        # If PLC does not supports this register format abort this generation and log the error
        if(registerFormat in ['uint64','int64','float64']):
            iecommon.logError('ERROR: In register '+registerName+', '+registerFormat+' format not supported for current controller model.', True)
            return
    
        #register has relative address irrespective of the protocol mode.
        registerOffset = long(registerDOM.prop('address'))
        iecommon.logDebug("Processing: %s %s %s %s" %(className, deviceLabel, blockName, registerName))

        # Compute the register address relying on the Class Parameters file data
        # Attention! Beckhoff uses WORD addressing while Parameters data are expressed in bytes to be PLC independent
        if deploy.plcModel == 'BC9020':
            registerAddress = (blockAddress - long(deploy.address) + (deviceIndex * blockMemSize) + registerOffset)/2
        elif deploy.plcModel == 'CX9020':
            registerAddress = (blockAddress - long(deploy.address) + (deviceIndex * blockMemSize) + registerOffset)
        else:
            raise "PLC model not supported: " + deploy.plcModel

        # add comment to source to identify block
        source += '    (*'+className+'/'+deviceLabel+'/'+blockName+' *)\r\n'
        
        # Compute the base checksum of the class name
        classNameCRC = zlib.crc32(className,0) & 0xffff
        
        if blockName == 'hdrBlk':
            if registerName == '_version':
                source+=beckhoffRegInit %(registerName, classNameCRC, deviceLabel, registerAddress, whichTwincatFormat[registerFormat]+"(16)", deploy.silecsVersion )
            if registerName == '_user':
                source+=beckhoffRegInit %(registerName, classNameCRC, deviceLabel, registerAddress, whichTwincatFormat[registerFormat]+"(16)", "'"+deploy.owner+"'") 
            if registerName == '_checksum':
                source+=beckhoffRegInit %(registerName, classNameCRC, deviceLabel, registerAddress, whichTwincatFormat[registerFormat], deploy.checksum)
            if registerName == '_date':
                source+=beckhoffRegInit %(registerName, classNameCRC, deviceLabel, registerAddress, whichTwincatFormat[registerFormat], getDateTime())
        
        else:   # not header block
             # set data type - for string with fixed value registerLength
            if registerFormat == 'string':
                format = whichTwincatFormat[registerFormat]+'('+str(registerLength)+')'
            else:
                format = whichTwincatFormat[registerFormat]
                
            if registerDimension1 == 1 and registerDimension2 == 1:    # scalar
                source += beckhoffReg %(registerName, classNameCRC, deviceLabel, registerAddress, format)
            elif registerDimension1 > 1 and registerDimension2 == 1: # array
                source += beckhoffRegArray %(registerName, classNameCRC, deviceLabel, registerAddress, registerDimension1-1, format)
            else:   # dim1>=1 and for whatever dim2, use double array syntax
                source += beckhoffRegArray2d %(registerName, classNameCRC, deviceLabel, registerAddress, registerDimension1-1, registerDimension2-1, format)
                    
    return source
# ----------------------------------------------------
def generateBeckhoffSources(paramDOM, sourceFolderPath, logTopics):
    deploy = getDeploymentInformation(paramDOM)
    
    # Prepare the global variables source (<.exp> file) with its diagnostic data-block header
    source ="""(* +-------------------------------------------------------------------\r\n"""
    source += """* | C.E.R.N Geneva, Switzerland\r\n"""
    source += """* | SILECS - BE/CO-FE\r\n"""
    source += """* | April 2015\r\n"""
    source += """* +-------------------------------------------------------------------\r\n"""
    source += """*\r\n"""
    source += """* Release : SILECS_%s\r\n"""%(deploy.silecsVersion)
    source += """*)"""
        
    source += '\r\nVAR_GLOBAL\r\n'

    classNodes = paramDOM.xpathEval("/SILECS-Param/SILECS-Mapping/SILECS-Class")
    for classNode in classNodes:
        address = int(classNode.prop('address')) # address of the first register of the configuration
        blockDOMList = classNode.xpathEval("Block") # get Blocks elements of that class
        deviceDOMList = classNode.xpathEval("Instance") # get Device instances of that class
            
        # Device mode is not supported for Beckhoff PLCs - Only block mode available
        for blockDOM in blockDOMList:
            for deviceIndex, deviceDOM in enumerate(deviceDOMList):
                source += generateBeckhoffRegisters(classNode, deviceDOM, blockDOM, deviceIndex,deploy,logTopics)
                
    source += 'END_VAR'

    # Write the source into the EXP source file
    generateControllerFiles(sourceFolderPath,deploy,".exp",source,logTopics);

#-------------------------------------------------------------------------
# NI code generation
#-------------------------------------------------------------------------
def generateNISources(paramDOM, sourceFolderPath ,logTopics):
    iecommon.logInfo("PLC source generation not available for National Instruments equipment",logTopics)
    return

def getDeploymentInformation(paramDOM):
    deploy = Deploy()
    mapping = paramDOM.xpathEval("/SILECS-Param/SILECS-Mapping")[0] 

    deploy.plcModel    = mapping.prop('plc-model')
    deploy.plcSystem    = mapping.prop('plc-system')
    deploy.plcProtocol = mapping.prop('protocol')
    deploy.plcName = mapping.prop('plc-name')
    deploy.address = mapping.prop('address')
    ownerNode = paramDOM.xpathEval("/SILECS-Param/Mapping-Info/Owner")[0] 
    deploy.owner = ownerNode.prop('user-login')
    deploymentNode = paramDOM.xpathEval("/SILECS-Param/Mapping-Info/Deployment")[0] 
    deploy.checksum = long(deploymentNode.prop('checksum'))
    rootNode = paramDOM.xpathEval("/SILECS-Param")[0] 
    deploy.silecsVersion = rootNode.prop("silecs-version")
    return deploy

def generateControllerFiles(sourceFolderPath,deploy,fileExtention,source,logTopics={'errorlog': True}):
    fileName = deploy.plcName + fileExtention
    fullFilePath = os.path.normpath(sourceFolderPath + "/" + fileName )

    iecommon.logInfo("Generate PLC sources: %s" %fullFilePath,logTopics);
    fdesc = open(fullFilePath, "w")
    fdesc.write(source)
    fdesc.close()

# Needed for unit-testing
def generateControllerCode(controllerNode, paramsFile, sourceFolderPath, logTopics={'errorlog': True}):
    paramDOM = libxml2.parseFile(paramsFile)
    plcNode  = controllerNode.xpathEval("*[@system]")[0]
    if plcNode.get_name() == 'Siemens-PLC':
        generateSiemensSources(paramDOM,sourceFolderPath,logTopics)
    elif plcNode.get_name() == 'Virtual-Controller':
        generateVirtualS7Sources(paramDOM,sourceFolderPath,logTopics)
    elif plcNode.get_name() == 'Schneider-PLC':
        generateSchneiderSources(paramDOM,sourceFolderPath,logTopics)
    elif plcNode.get_name() == 'Rabbit-uC':
        generateRabbitSources(paramDOM,sourceFolderPath,logTopics)
    elif plcNode.get_name() == 'Beckhoff-PLC':
        generateBeckhoffSources(paramDOM,sourceFolderPath,logTopics)
    elif plcNode.get_name() == 'NI-Controller':
        generateNISources(paramDOM,sourceFolderPath,logTopics)
    else:
        iecommon.logError("The PLC-Type %s is not supported" %plcNode.get_name(), True,logTopics)
    paramDOM.freeDoc()

#=========================================================================
# Entry Point
#=========================================================================
def genPlcSrc(workspacePath, deployName,silecsVersion,logTopics={'errorlog': True}):
    deployFilePath = iefiles.getSilecsDeployFilePath(workspacePath,deployName)
    deployDOM = libxml2.parseFile(deployFilePath)
    iecommon.logInfo("Load Deployment document: %s" %deployFilePath,logTopics);

    # Create the source folder if needed
    sourceFolderPath = iefiles.getControllerSourcesDirectory(workspacePath, deployName)
    if not os.path.exists(sourceFolderPath):
        os.makedirs(sourceFolderPath)

    # Add xml of the silecs-header
    iecommon.addSilecsHeaderToClasses(deployDOM)

    controllerNodes = deployDOM.xpathEval("/SILECS-Deploy/Controller")
    for controllerNode in controllerNodes:
        paramsFile = iefiles.getParameterFile(workspacePath, deployName,controllerNode.prop("host-name"))
        generateControllerCode(controllerNode,paramsFile,sourceFolderPath,logTopics)

    # Clean-up the memory before exit
    deployDOM.freeDoc()
    
    return "Genplcsrc succeded!"
 
