// Copyright 2016 CERN and GSI
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "PLCRegister.h"

namespace Silecs
{

	PLCRegister::PLCRegister(Device* theDevice,const boost::shared_ptr<ElementXML>& registerNode) : Register(theDevice, registerNode) {}
	PLCRegister::~PLCRegister() {}

	BigEndianRegister::BigEndianRegister(Device* theDevice,const boost::shared_ptr<ElementXML>& registerNode) : PLCRegister(theDevice, registerNode)  {}
    BigEndianRegister::~BigEndianRegister() {}

	LittleEndianRegister::LittleEndianRegister(Device* theDevice,const boost::shared_ptr<ElementXML>& registerNode) : PLCRegister(theDevice, registerNode) {}
	LittleEndianRegister::~LittleEndianRegister() {}

	S7Register::S7Register(Device* theDevice,const boost::shared_ptr<ElementXML>& registerNode) : BigEndianRegister(theDevice, registerNode) {}
    S7Register::~S7Register() {}

	UnityRegister::UnityRegister(Device* theDevice,const boost::shared_ptr<ElementXML>& registerNode) : LittleEndianRegister(theDevice, registerNode) {}
	UnityRegister::~UnityRegister() {}

	TwinCATRegister::TwinCATRegister(Device* theDevice,const boost::shared_ptr<ElementXML>& registerNode) : UnityRegister(theDevice, registerNode) {}
	TwinCATRegister::~TwinCATRegister() {}


    /// Generic methods for byte-swapping ----------------------------------------------
    uint16_t PLCRegister::_swaps(uint16_t s)
    {
        return ( ((s & 0x00ff) << 8) + ((s & 0xff00) >> 8) );
    }

    uint32_t PLCRegister::_swapl(uint32_t l)
    {
        uint32_t ul;
        void *pl = &l;
        void *pul = &ul;
        *((uint16_t *)pul)   = _swaps(*((uint16_t *)pl+1));
        *((uint16_t *)pul+1) = _swaps(*((uint16_t *)pl));
        return(ul);
    }

    //Beckhoff/BC9020 PLC uses the same memory swapping than Unity PLCs except for float32 data
    //This method is used to specialized the 32bits data swapping inherited from the UnityRegister class.
    uint32_t TwinCATRegister::_swapl(uint32_t l)
    {
        uint32_t ul;
        void *pl = &l;
        void *pul = &ul;

        if ( (getPLC()->getModel() == "BC9020") && (getFormat() == Silecs::Float32) )
        {  //Particular case for BC9020 PLC:  swap only bytes together (not the words)
            *((uint16_t *)pul)   = _swaps(*((uint16_t *)pl));
            *((uint16_t *)pul+1) = _swaps(*((uint16_t *)pl+1));
        }
        else
        { //Same as base-class methods for other controllers
            ul = PLCRegister::_swapl(l);
        }
        return(ul);
    }

    float PLCRegister::_swapf(float f)
    {
        uint32_t l;
        void *pl = &l;
        void *pf = &f;
        l = _swapl(*((uint32_t *)pf));
        return(*((float *)pl));
    }

    double PLCRegister::_swapd(double d)
    {
        uint64_t ld;
        void *pd = &d;
        void *pld = &ld;
        *((uint32_t *)pld)   = _swapl(*((uint32_t *)pd+1));
        *((uint32_t *)pld+1) = _swapl(*((uint32_t *)pd));
        return(*((double *)pld));
    }


    // BIG-ENDIANNESS  ---------------------------------------------------------------------------
    void BigEndianRegister::swapBytesNToH(unsigned char *data, unsigned long size)
    {
        switch(size)
        {
            case 1:
                //for CHAR and BYTE, nothing to do in that case.
                break;
            case 2: {
                //WORD or INT type
                uint16_t s = ntohs(*(uint16_t *)data);
                //no swap, just convert to network format
                *(uint16_t *)data = s;
                break;
            }
            case 4: {
                //DWORD, DINT or REAL type
                uint32_t l = ntohl(*(uint32_t *)data);
                //no swap, just convert to network format
                *(uint32_t *)data = l;
                break;
            }
    #ifdef __x86_64__
            case 8: {
                //uInt64, Int64 or Float64
                //direct conversion from little-endian to host exists only from SLC6 version
                uint64_t l64 = le64toh(*(uint64_t *)data);
                *(uint64_t *)data = l64;
                break;
            }
    #endif
            default:
                //nothing to do unknown type
                break;
        }
    }

    void BigEndianRegister::swapBytesHToN(unsigned char *data, unsigned long size)
    {
        switch(size)
        {
            case 1:
                //for CHAR and BYTE, nothing to do in that case.
                break;
            case 2: {
                //WORD or INT type
                uint16_t s = htons(*(uint16_t *)data);
                //no swap, just convert to network format
                *(uint16_t *)data = s;
                break;
            }
            case 4: {
                //DWORD, DINT or REAL type
                uint32_t l = htonl(*(uint32_t *)data);
                //no swap, just convert to network format
                *(uint32_t *)data = l;
                break;
            }
    #ifdef __x86_64__
            case 8: {
                //uInt64, Int64 or Float64
                //direct conversion from host to little-endian exists only from SLC6 version
                uint64_t l64 = htole64(*(uint64_t *)data);
                *(uint64_t *)data = l64;
                break;
            }
    #endif
            default:
                //nothing to do unknown type
                break;
          }
    }

    // LITTLE-ENDIANNESS  ---------------------------------------------------------------------------
    void LittleEndianRegister::swapBytesNToH(unsigned char *data, unsigned long size)
    {
        switch(size)
        {
            case 1:
                //for CHAR and BYTE, nothing to do in that case.
                break;
            case 2: {
                //WORD or INT type
                uint16_t s = ntohs(*(uint16_t *)data);
                  //convert to network format and word swapping
                  *(uint16_t *)data = _swaps(s);
                break;
            }
            case 4: {
                //DWORD, DINT or REAL type
                uint32_t l = ntohl(*(uint32_t *)data);
                //convert to network format and word swapping
                *(uint32_t *)data = _swapl(l);
                break;
            }
            #ifdef __x86_64__
            case 8: {
                //uInt64, Int64 or Float64
                //direct conversion from big-endian to host exists only from SLC6 version
                uint64_t l64 = be64toh(*(uint64_t *)data);
                *(uint64_t *)data = l64;
                break;
            }
            #endif
            default:
                //nothing to do unknown type
                break;
            }
    }

    void LittleEndianRegister::swapBytesHToN(unsigned char *data, unsigned long size)
    {
        switch(size)
        {
            case 1:
                //for CHAR and BYTE, nothing to do in that case.
                break;
            case 2: {
                //WORD or INT type
                uint16_t s = htons(*(uint16_t *)data);
                //convert to network format and word swapping
                *(uint16_t *)data = _swaps(s);
                break;
            }
            case 4: {
                //DWORD, DINT or REAL type
                uint32_t l = htonl(*(uint32_t *)data);
                //convert to network format and word swapping
                *(uint32_t *)data = _swapl(l);
                break;
            }
    #ifdef __x86_64__
            case 8: {
                //uInt64, Int64 or Float64
                //direct conversion from host to big-endian exists only from SLC6 version
                uint64_t l64 = htobe64(*(uint64_t *)data);
                *(uint64_t *)data = l64;
                break;
            }
    #endif
            default:
                //nothing to do unknown type
                break;
          }
    }


    // ----------------------------------------------------------------------------------------
	void PLCRegister::importValue(void* pBuffer, timeval ts)
	{
		//Compute the register address within the block: block-address + register-offset
		unsigned char* pDataBase = ((unsigned char*)pBuffer) + address_;
		unsigned char* pData = pDataBase;

		if ((format_ == Int8) || (format_ == uInt8))
		{
			//Some PLC use more than 1 byte for this format (exp.: SCHNEIDER aligns char on 16bits)
		    unsigned int step = memSize_ / dimension1_ / dimension2_ / size_;
		    /*TODO:
			 * Optimization a tester: utiliser un memcpy si (step == 1) au lieu du for
			 * (idem pour l'export)
			 */
		    for (unsigned int i=0; i<(dimension1_); i++)
            {
                for (unsigned int j=0; j<(dimension2_); j++)    // works also for 1d, in this case j is always 0 and points to the element on the array
                {
                    // compact the array by removing the alignment added
                    ((unsigned char*)pRecvValue_)[(i*dimension2_)+j] = pData[(i*step*dimension2_)+j];
                }
            }
		}
		else
		{
			if (format_ != Date)
			{ 	//Short, uShort, Long, uLong, Float
			    for (unsigned int i=0; i<(dimension1_*dimension2_); i++)
				{
				    swapBytesNToH(pData, size_);
				    pData += size_;
				}

				//The buffer is correctly aligned and swapped now. We can transfer the value to the register cell.
				memcpy(pRecvValue_, (void*)pDataBase, (size_*dimension1_*dimension2_));
			}
			else
			{
			    //Date&Time
                unsigned long dtSize = size_; //Most of controllers use 64bits coding for DT type
                                             //Beckhoff controller uses 32bits (see below)

			    for (unsigned int i=0; i<dimension1_; i++)
                {
                    for (unsigned int j=0; j<dimension2_; j++)    // works also for 1d, in this case j is always 0 and points to the element on the array
                    {
                        switch(getPLC()->getBrandID())
                        {
                            case Siemens:
                                ((double *)pRecvValue_)[i*dimension2_+j] = IeRfcGetTime(pData);
                                break;
                            case Schneider:
                            case Digi:
                                ((double *)pRecvValue_)[i*dimension2_+j] = IeMdbGetTime(pData);
                                break;
                            case Ni:
                                //TODO
                                break;
                            case Beckhoff:
                                //Beckhoff controller implement the standard EPOCH on 32bits word
                                dtSize = sizeof(uint32_t);
                                uint32_t *timeVal = (uint32_t *)pData;
                                swapBytesNToH((unsigned char *)&(timeVal[0]), sizeof(uint32_t));
                                ((double *)pRecvValue_)[i*dimension2_+j] = static_cast<double>(timeVal[0]);
                                break;
                        }
                        pData += dtSize;
                    }
                }
			}
		}

		//Set the register time-stamp
		tod_ = ts;

		//DATA topic makes sense with RECV one
		if (RECV & Log::topics_) LOG(DATA) << theDevice_->getPLC()->getName() << "/" << theDevice_->getLabel() << "/" << getName() << ", len: " << getLength() << ", dim1: " << getDimension1() << ", dim2: " << getDimension2() << ", ts: " << getTimeStampAsString() <<  ", value: " << getValAsByteString(pRecvValue_, 10);
	}


	void PLCRegister::exportValue(void* pBuffer)
	{
		//DATA topic makes sense with SEND one
		if (SEND & Log::topics_) LOG(DATA) << theDevice_->getPLC()->getName() << "/" << theDevice_->getLabel() << "/" << getName() << ", len: " << getLength() << ", dim1: " << getDimension1() << ", dim2: " << getDimension2() <<  ", value: " << getValAsByteString(pSendValue_, 10);

		//Compute the register address within the block: block-address + register-offset
		unsigned char* pData = ((unsigned char*)pBuffer) + address_;

		if ((format_ == Int8) || (format_ == uInt8))
		{
			//Some PLC use more than 1 byte for this format (exp.: SCHNEIDER aligns char on 16bits)
			unsigned int step = memSize_ / dimension1_ / dimension2_ / size_;

			for (unsigned int i=0; i<dimension1_; i++)
            {
                for (unsigned int j=0; j<(dimension2_); j++)
                {
                    // expand the array by adding required space
                    pData[(i*step*dimension2_)+j] = ((unsigned char*)pSendValue_)[(i*dimension2_)+j];
                }
            }
		}
		else
		{
			if (format_ != Date)
			{
				//Short, uShort, Long, uLong, Float
				//Transfert the value of the register cell to the buffer
				memcpy((void*)pData, pSendValue_, (size_*dimension1_*dimension2_));

				for (unsigned int i=0; i<(dimension1_*dimension2_); i++)
				{
				    swapBytesHToN(pData, size_);
				    pData += size_;
				}
			}
			else
			{ 	//Date&Time
			    unsigned long dtSize = size_; //Most of controllers use 64bits coding for DT type
			                                 //Beckhoff controller uses 32bits (see below)

                for (unsigned int i=0; i<(dimension1_*dimension2_); i++)
                {
                    time_t epoch = static_cast<time_t>((double)((double *)pSendValue_)[i]);
                    switch(getPLC()->getBrandID())
                    {
                        case Siemens:
                            IeRfcSetTime(pData, epoch);
                            break;
                        case Schneider:
                        case Digi:
                            IeMdbSetTime(pData, epoch);
                            break;
                        case Ni:
                            //TODO
                            break;
                        case Beckhoff:
                            //Beckhoff controller implement the standard EPOCH on 32bits word
                            dtSize = sizeof(uint32_t);
                            uint32_t *timeVal = (uint32_t *)pData;
                            timeVal[0] = static_cast<uint32_t>(epoch);
                            swapBytesHToN((unsigned char *)&(timeVal[0]), sizeof(uint32_t));
                            break;
                    }
                    pData += dtSize;
                }
			}
		}
	}

	void PLCRegister::copyValue()
	{
	    //Transfer the value of the recvBuffer to to the sendBuffer
	    if (format_ == String)
        {
	        std::string** pRecvStringValue_ = static_cast<std::string**>(pRecvValue_);
	        std::string** pSendStringValue_ = static_cast<std::string**>(pSendValue_);
	        for(unsigned int i=0; i<(dimension1_*dimension2_); i++)
            {
	            *(pSendStringValue_[i]) = *(pRecvStringValue_[i]);
            }
        }
	    else
	    {
	        memcpy(pSendValue_, pRecvValue_, (size_*dimension1_*dimension2_));
	    }
	}


    // S7 METHODS -------------------------------------------------------------------------------
	void S7Register::importString(void* pBuffer, timeval ts)
	{
        // Compute the register address within the block: block-address + register-offset
        unsigned char* pDataBase = ((unsigned char*)pBuffer) + address_;
        std::string** pRecvStringValue_ = static_cast<std::string**>(pRecvValue_);

        //Some PLC use more than 1 byte for this format (exp.: SIEMENS aligns string on 16bits address)
        unsigned int step = memSize_ / dimension1_ / dimension2_ / size_;

        // for each string of the single/double array -
        for (unsigned long i=0; i<(dimension1_*dimension2_); i++) // scalar has dim1=dim2=1
        {    unsigned char* pData = &(pDataBase[i*step]);
            pRecvStringValue_[i]->assign((const char*)&(pData[2]), (size_t)pData[1]);
        }
        //Set the register time-stamp
        tod_ = ts;

        //DATA topic makes sense with RECV one
        if (RECV & Log::topics_) LOG(DATA) << theDevice_->getPLC()->getName() << "/" << theDevice_->getLabel() << "/" << getName() << ", len: " << getLength() << ", dim1: " << getDimension1() << ", dim2: " << getDimension2() << ", ts: " << getTimeStampAsString() <<  ", value: " << getValAsByteString(pRecvValue_, 10);
	}

	void S7Register::exportString(void* pBuffer)
	{
        //DATA topic makes sense with SEND one
        if (SEND & Log::topics_) LOG(DATA) << theDevice_->getPLC()->getName() << "/" << theDevice_->getLabel() << "/" << getName() << ", len: " << getLength() << ", dim1: " << getDimension1() << ", dim2: " << getDimension2() <<  ", value: " << getValAsByteString(pSendValue_, 10);

        //Compute the register address within the block: block-address + register-offset
        unsigned char* pData = ((unsigned char*)pBuffer) + address_;
        std::string** pSendStringValue_ = static_cast<std::string**>(pSendValue_);

        //Some PLC use more than 1 byte for this format (exp.: SIEMENS aligns string on 16bits address)
        unsigned int step = memSize_ / dimension1_ / dimension2_ / size_;

        // for each string of the single/double array -
        for (unsigned long i=0; i<dimension1_*dimension2_; i++)  // scalar has dim1=dim2=1
        {   //Current size of the sent string should not exceed the register string max length
            std::string sendStringValue = *(pSendStringValue_[i]);
            if (sendStringValue.size() > length_) sendStringValue = sendStringValue.substr(0, length_);
            // S7 requires that the first two bytes of string are reserved for length and max length of string
            pData[(i*step)+0] = (unsigned char)length_ & 0XFF;    // convert to byte (max length is 254 (+2) = 1 byte)
            pData[(i*step)+1] = (unsigned char)sendStringValue.size() & 0XFF;
            memcpy((void *)&(pData[(i*step)+2]), (void *)sendStringValue.c_str(), sendStringValue.size());
        }
	}

    // UNITY METHODS -----------------------------------------------------------------------------------
	// TwinCAT object inherit from the same methods than UNITY.
	void UnityRegister::importString(void* pBuffer, timeval ts)
    {
        // Compute the register address within the block: block-address + register-offset
        unsigned char* pDataBase = ((unsigned char*)pBuffer) + address_;
        std::string** pRecvStringValue_ = static_cast<std::string**>(pRecvValue_);

        //Some PLC use more than 1 byte for this format (exp.: SCHNEIDER aligns char on 16bits)
        unsigned int step = memSize_ / dimension1_ / dimension2_ / size_;

        // for each string of the single/double array -
        for (unsigned long i=0; i<(dimension1_*dimension2_); i++) // scalar has dim1=dim2=1
        {   unsigned char* pData = &(pDataBase[i*step]);
            size_t strSize =  min((size_t)length_, strlen((const char*)pData));
            pRecvStringValue_[i]->assign((const char*)pData, strSize);
        }
        tod_ = ts;

        //DATA topic makes sense with RECV one
        if (RECV & Log::topics_) LOG(DATA) << theDevice_->getPLC()->getName() << "/" << theDevice_->getLabel() << "/" << getName() << ", len: " << getLength() << ", dim1: " << getDimension1() << ", dim2: " << getDimension2() << ", ts: " << getTimeStampAsString() <<  ", value: " << getValAsByteString(pRecvValue_, 10);
    }

	void UnityRegister::exportString(void* pBuffer)
	{
        //DATA topic makes sense with SEND one
        if (SEND & Log::topics_) LOG(DATA) << theDevice_->getPLC()->getName() << "/" << theDevice_->getLabel() << "/" << getName() << ", len: " << getLength() << ", dim1: " << getDimension1() << ", dim2: " << getDimension2() <<  ", value: " << getValAsByteString(pSendValue_, 10);

        //Compute the register address within the block: block-address + register-offset
        unsigned char* pData = ((unsigned char*)pBuffer) + address_;
        std::string** pSendStringValue_ = static_cast<std::string**>(pSendValue_);

        // Some PLC use more than 1 byte for this format (exp.: SCHNEIDER aligns char on 16bits)
        unsigned int step = memSize_ / dimension1_ / dimension2_ / size_;

        // for each string of the single/double array -
        for (unsigned long i=0; i<dimension1_*dimension2_; i++)  // scalar has dim1=dim2=1
        {   //Current size of the sent string should not exceed the register string max length
            size_t strSize = pSendStringValue_[i]->size();
            std::string sendStringValue = pSendStringValue_[i]->substr(0, min((size_t)length_, strSize));
            bzero((void *)&(pData[i*step]), step);  //TwinCAT string requires '\0' terminator (mem-size of the string (~step) has been properly dimensioned in the mapping).
            memcpy((void *)&(pData[i*step]), (void *)sendStringValue.c_str(), sendStringValue.size());
        }
	}

} // namespace

