// Copyright 2016 CERN and GSI
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <silecs-communication/interface/core/SilecsService.h>
#include <silecs-communication/interface/utility/Thread.h>
#include <silecs-communication/interface/utility/XMLParser.h>
#include <silecs-communication/interface/equipment/SilecsPLC.h>
#include <silecs-communication/interface/equipment/SilecsBlock.h>
#include <silecs-communication/interface/equipment/SilecsRegister.h>
#include <silecs-communication/interface/core/SilecsAction.h>
#include <silecs-communication/interface/core/WrapperAction.h>
#include <silecs-communication/interface/utility/SilecsException.h>
#include <silecs-communication/interface/utility/SilecsLog.h>
#include <silecs-communication/interface/utility/StringUtilities.h>

namespace Silecs
{

	Block::Block(PLC* thePLC, ElementXML blockNode, AccessType accessType) :
			thePLC_ (thePLC), hasMaster_(false), hasSlave_(false)
	{
		std::string designName = blockNode.getAttribute("name");

        name_ = designName;
		StringUtilities::fromString(size_,blockNode.getAttribute("size"));
		StringUtilities::fromString(memSize_,blockNode.getAttribute("mem-size"));
		StringUtilities::fromString(address_,blockNode.getAttribute("address"));

        resetCustomAttributes(); //by default custom-size of the block is the maximum size

        /*In case of InOut block we will instantiate 2 different blocks (input/output)
         * and the accessType is given by the caller.
		*/
        accessType_ = accessType;

        // Adjust the synchro-type of the block relying on its own registers
        // . hasMaster if the block contains 1 MASTER register at least
        // . hasSlave if the block contains 1 SLAVE register at least
        // can be a mix or none if only Volatile registers
		std::vector< boost::shared_ptr<ElementXML> > registerNodes = blockNode.childList_;
		std::vector< boost::shared_ptr<ElementXML> >::const_iterator registerIter;
		for(registerIter = registerNodes.begin(); registerIter != registerNodes.end(); registerIter++)
		{
			std::string synchro = (*registerIter)->getAttribute("synchro");
			if (!hasMaster_ && Service::whichSynchroType(synchro) == Master)
				hasMaster_ = true;
			if (!hasSlave_ && Service::whichSynchroType(synchro) == Slave)
				hasSlave_ = true;
		}
	}

	Block::~Block()
	{
		if (DEBUG & Log::topics_) LOG(ALLOC) << "Block (delete): " << name_;

		if (pAction_ != NULL) delete pAction_;
		if (pTask_ != NULL) delete pTask_;
	}


	AccessType Block::whichAccessType(std::string type)
	{
		StringUtilities::toLower(type);
		if (type == "write-only") return Output;
		else if (type == "read-only") return Input;
		else if (type == "read-write") return InOut;
		else throw SilecsException(__FILE__, __LINE__, DATA_UNKNOWN_ACCESS_TYPE, type);
	}


	std::string Block::whichAccessType(AccessType type)
	{
		switch(type)
		{
		case Output: return "WRITE-ONLY"; break;
		case Input: return "READ-ONLY"; break;
		case InOut: return "READ-WRITE"; break;
		default: throw SilecsException(__FILE__, __LINE__, DATA_UNKNOWN_ACCESS_TYPE, StringUtilities::toString(type));
		}
	}

    void Block::setCustomAttributes(const unsigned long customAddress, const unsigned long customOffset, const unsigned long customSize)
    {
        //user wants to limit the size of data to be sent/receive - check that it is lower than design allocation size
        if (customSize > memSize_)
            throw SilecsException(__FILE__, __LINE__, PARAM_EXCEEDED_BLOCK_SIZE, StringUtilities::toString(memSize_));

        customAddress_      = customAddress;
        customOffset_       = customOffset;
        customSize_         = customSize;
        customAttributes_   = true;
        LOG(DEBUG) << "Set custom attributes of block: " << name_ << ", memSize: " << memSize_ << ", customSize: " << customSize << ", customAddress: " << customAddress << ", customOffset: " << customOffset;
    }

    void Block::resetCustomAttributes()
    {
        customAttributes_ = false;
        LOG(DEBUG) << "Reset custom attributes of block: " << name_;
    }


} // namespace
