// Copyright 2016 CERN and GSI
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <silecs-communication/interface/equipment/SilecsCluster.h>
#include <silecs-communication/interface/equipment/SilecsPLC.h>
#include <silecs-communication/interface/equipment/SilecsDevice.h>
#include <silecs-communication/interface/equipment/SilecsBlock.h>
#include <silecs-communication/interface/utility/SilecsException.h>
#include <silecs-communication/interface/utility/SilecsLog.h>
#include <silecs-communication/protocol/core/silecs.h>

namespace Silecs
{
    bool ClusterConfig::getSharedConnection() const
    {
        return sharedConnection;
    }

    void ClusterConfig::setSharedConnection(bool sharedConnectionFlag)
    {
        sharedConnection = sharedConnectionFlag;
    }


	Cluster::Cluster(const std::string className, const std::string classVersion) : className_ (className), classVersion_ (classVersion)
	{
		char hostName[128];
	    if (gethostname(hostName, 128) == -1)
			throw SilecsException(__FILE__, __LINE__, errno, strerror(errno));

	    //store the hostname by removing the domain-name (.cern.ch) if any.
	    hostName_ = std::string(hostName);
	    removeDomainName(hostName_);

	    //Default run-time configuration setup
	    configuration.sharedConnection = false;  //Do not shared connection with other clusters by default

		LOG((ALLOC|DIAG)) << "CLUSTER (create): " << className_ << "/" << classVersion_;
	}

	Cluster::~Cluster()
	{
		LOG(ALLOC) << "CLUSTER (delete): " << className_ << "/" << classVersion_;

		// Remove the PLCs instances related to this Cluster (hostName_/className_)
		plcMapType::iterator plcMapIter;
		for(plcMapIter = plcMap_.begin(); plcMapIter != plcMap_.end(); ++plcMapIter)
		{ delete plcMapIter->second;
		}
	    plcMap_.clear();
	    IPAddrMap_.clear(); //this map refer the same PLC object (not necessary to delete them)
	}

	void Cluster::removeDomainName(std::string& hostName)
	{   size_t domain = hostName.find_first_of('.');
		if (domain != std::string::npos) hostName.erase(domain);
	}

	bool Cluster::getHostNameByAddress(const std::string IPAddress, std::string& hostName)
	{
		struct hostent *he = NULL;
		struct in_addr ipv4addr;

		if (inet_pton(AF_INET, IPAddress.c_str(), &ipv4addr) == 1)
		{   he = gethostbyaddr(&ipv4addr, sizeof ipv4addr, AF_INET);
		}

		if (he)
		{	hostName = std::string(he->h_name);
			if (DEBUG & Log::topics_) LOG(COMM) << "getHostNameByAddress()  IPaddr: " << IPAddress << ", host-name: " << hostName;
			return true;
		}
		return false;
	}

	bool Cluster::getAddressByHostName(const std::string hostName, std::string& IPAddress)
	{
		struct in_addr addr;
		struct hostent *hp;
		char *ipstr;

		hp = gethostbyname(hostName.c_str());
		if (hp)
		{
			addr.s_addr = *((unsigned long int *) hp->h_addr);
			ipstr = inet_ntoa(addr);
			IPAddress = std::string(ipstr);
			if (DEBUG & Log::topics_) LOG(COMM) << "getAddressByHostName()  host-name: " <<  hostName << ", IPaddr: " << IPAddress;
			return true;
		}
		return false;
	}


	Block* Cluster::getBlock(const std::string blockName, AccessType accessType)
	{	// Retrieve the block from the first PLC (same for each of them)
		return plcMap_.begin()->second->getBlock(blockName, accessType);
	}


	std::string Cluster::getBlockList(AccessType accessType)
	{
		std::string blockList = "";
		//At first, check this class/version is already deployed (one PLC in the Cluster at least)
		if (plcMap_.size() > 0)
		{	// Retrieve the block-map from the first PLC (same for each of them)
			blockVectorType blockCol = plcMap_.begin()->second->getBlockCol();

			blockVectorType::iterator block;
			for(block = blockCol.begin(); block != blockCol.end(); ++block)
			{
				if ((*block)->getAccessType() == accessType)
				{
					std::string blockName = (*block)->getName();
					blockList.append(blockName + " ");
				}
			}
		}
		return blockList;
	}

	PLC* Cluster::getPLC(std::string plcID, string parameterFile)
	{
		//getPLC() supports IP and host-name addressing, but both are required to register the PLC object:
		//Host-name will be used to manage related configuration documents.
		//IP-addr is used to connect PLC in order to reduce the naming-server accesses.

        //Force PLC hostname lower-case
        StringUtilities::toLower(plcID);

		//At first, check if the given ID is already registered in that Cluster (host-name or IP-address)
		bool plcNotFound = false;
		plcMapType::iterator iter = plcMap_.find(plcID);   //check the host-name map at first
		if (iter == plcMap_.end())
		{	//we didn't find PLC by host-name, try to find it by IP address
			iter = IPAddrMap_.find(plcID);
			plcNotFound = (iter == IPAddrMap_.end());
		}

		if (plcNotFound)
		{
			//Retrieve Host-name & IP-address of the PLC
			std::string hostName;
			std::string IPAddress;

			if (isdigit(plcID[0]))
			{   //user provided an IP address (start with digit character)
				IPAddress = plcID;
				if (!getHostNameByAddress(IPAddress, hostName))
					throw SilecsException(__FILE__, __LINE__, PARAM_UNKNOWN_IP_ADDRESS, IPAddress);
			}
			else
			{	//user provided an PLC network name (start with alpha-character)
				hostName = plcID;
				if (!getAddressByHostName(hostName, IPAddress))
					throw SilecsException(__FILE__, __LINE__, PARAM_UNKNOWN_PLC_HOSTNAME, hostName);
			}
			removeDomainName(hostName); //remove ".cern.ch" if any

			//Then create the new instance of the PLC

			PLC* pPLC = new PLC(this, hostName, IPAddress, parameterFile);
			plcMap_.insert(std::make_pair(hostName, pPLC));      //update host-name map
			IPAddrMap_.insert(std::make_pair(IPAddress, pPLC));  //update IP address map
			return pPLC;
		}
		return iter->second;
	}


	std::string Cluster::getHostName() { return hostName_; }
	std::string Cluster::getClassName() { return className_; }
	std::string Cluster::getClassVersion() { return classVersion_; }


	std::string Cluster::getPLCList()
	{
		std::string PLCList = "";
		plcMapType::iterator plcMapIter;
		for(plcMapIter = plcMap_.begin(); plcMapIter != plcMap_.end(); ++plcMapIter)
		{	if (!PLCList.empty()) PLCList.append(" ");
			PLCList.append(plcMapIter->second->getName());
		}
		return PLCList;
	}


	const plcMapType& Cluster::getPLCMap() { return plcMap_; }


	void Cluster::recv(std::string blockName)
	{
		//Asynchronous data receive
		plcMapType::iterator plcMapIter;

	    // Schedule task on each PLC using asynchronous call for parallelism
		for(plcMapIter = plcMap_.begin(); plcMapIter != plcMap_.end(); ++plcMapIter)
		{	plcMapIter->second->recvAsync(blockName);
		}

	    // then wait for the completion of each PLC thread (finally, maximum delay will be the longest one).
		for(plcMapIter = plcMap_.begin(); plcMapIter != plcMap_.end(); ++plcMapIter)
		{	plcMapIter->second->waitAsync();
		}
	}


	void Cluster::send(std::string blockName)
	{
		//Asynchronous data send
		plcMapType::iterator plcMapIter;

	    // Schedule task on each PLC using asynchronous call for parallelism
		for(plcMapIter = plcMap_.begin(); plcMapIter != plcMap_.end(); ++plcMapIter)
		{
			plcMapIter->second->sendAsync(blockName);
		}

	    // then wait for the completion of each PLC thread (finally, maximum delay will be the longest one).
		for(plcMapIter = plcMap_.begin(); plcMapIter != plcMap_.end(); ++plcMapIter)
		{
			plcMapIter->second->waitAsync();
		}
	}

	void Cluster::copyInToOut(const std::string blockName)
	{
		plcMapType::iterator plcMapIter;
		for(plcMapIter = plcMap_.begin(); plcMapIter != plcMap_.end(); ++plcMapIter)
			plcMapIter->second->copyInToOut(blockName);
	}


} // namespace

